/*
 * Resolve a hostname
 *
 * Copyright (C) 2003, SuSE Linux AG
 * Written by okir@suse.de
 */

#include <sys/poll.h>
#include <netdb.h>
#include <string.h>
#include <time.h>
#include <malloc.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>

#if defined(__GLIBC__)
#define dprintf mydprintf
#endif /* __GLIBC__ */

/* Number of times we retry resolving a name */
#define RESOLVER_RETRIES	10
/* Timeout between retries to resolve a name */
#define RESOLVER_RETRY_TIME	(3 * 60)
/* Time to wait for the sub-process comes up with an
 * answer, before we suspend the entry */
#define RESOLVER_WAIT_TIME	5

struct res_name {
	struct res_name *	next;
	char *			name;
	unsigned int		retries;
	time_t			retry_time;
	time_t			wait_time;
	pid_t			process;
	int			pipe;
	struct sockaddr_storage	address;
};

extern void	dprintf(char *fmt, ...);
static int	do_result(struct res_name *res);
static void	do_query(int, int, const char *);

static struct res_name *resolve;

int
resolve_name(const char *name, int family, struct sockaddr_storage *addr)
{
	struct res_name	*res;
	int		fd[2];
	pid_t		pid;
	time_t		now;

	time(&now);
	for (res = resolve; res; res = res->next) {
		if (!strcmp(res->name, name)) {
			/* If there's still a sub-process
			 * attached, see if it came up with an
			 * answer in the meantime */
			if (res->pipe >= 0)
				do_result(res);
			goto check_result;
		}
	}

	res = (struct res_name *) calloc(1, sizeof(*res));
	res->name = strdup(name);
	res->retries = RESOLVER_RETRIES;
	res->retry_time = now + RESOLVER_RETRY_TIME;
	res->pipe = -1;
	res->next = resolve;
	resolve = res;

retry:	res->retries -= 1;

	dprintf("Trying to resolve \"%s\", attempt #%u\n",
			res->name, RESOLVER_RETRIES - res->retries);

	if (res->pipe >= 0) {
		/* Make sure resolver subprocess is dead */
		kill(res->process, SIGKILL);
		close(res->pipe);
		res->pipe = -1;
	}

	signal(SIGCHLD, SIG_IGN);
	if (pipe(fd) < 0) {
		/* log error? */
		return -1;
	}
	if ((pid = fork()) < 0) {
		/* log error? */
		close(fd[0]);
		close(fd[1]);
		return -1;
	}

	if (pid == 0) {
		close(fd[0]);
		do_query(fd[1], family, res->name);
		/* shouldn't return */
		exit(1);
	}

	res->wait_time = now + RESOLVER_WAIT_TIME;
	res->process = pid;
	res->pipe = fd[0];
	close(fd[1]);

	/* Try to collect result within the next couple
	 * of seconds */
	do_result(res);

check_result:
	if (res->pipe >= 0) {
		dprintf("%s: name resolution not complete, "
			"waiting for resolver sub-process\n",
			res->name);
		return 0;
	}

	if (res->address.ss_family != AF_UNSPEC) {
		*addr = res->address;
		return 1;
	}
	if (res->retries == 0) {
		dprintf("%s: too many failed attempts to resolve hostname, "
			"given up.\n", res->name);
		return -1;
	}
	if (res->retry_time > now) {
		dprintf("%s: name resolution failed, "
			"will retry in %d seconds\n",
			res->name, res->retry_time - now);
		return 0;
	}
	goto retry;
}

/*
 * This function clears the resolver's internal state
 * when syslogd received a SIGHUP
 */
void
init_resolver(void)
{
	struct res_name	*res;

	for (res = resolve; res; res = res->next) {
		res->address.ss_family = AF_UNSPEC;
		res->retries = RESOLVER_RETRIES;
		res->retry_time = 0;
	}
}


/*
 * Collect result from sub-process
 */
int
do_result(struct res_name *res)
{
	struct sockaddr_storage addr;
	struct pollfd	_poll;
	long		wait;
	int		n;

	if (res->pipe < 0)
		return 0;

	/* See if we should wait for the subprocess to come
	 * up with an answer (the initial call to do_result
	 * after starting the sub-process will always wait for
	 * a few seconds to avoid losing messages) */
	if ((wait = res->wait_time - time(NULL)) < 0)
		wait = 0;

	/* See if we get the resolver's result within the
	 * next "wait" seconds. If not, let it continue in the
	 * background */
	_poll.fd = res->pipe;
	_poll.events = POLLIN;
	if (poll(&_poll, 1, wait * 1000) < 1)
		return 0;

	n = read(res->pipe, &addr, sizeof(addr));

	/* child process has done its duty. */
	kill(res->process, SIGKILL);
	close(res->pipe);
	res->process = 0;
	res->pipe = -1;

	if (n != sizeof(addr))
		return -1;

	res->address = addr;
	if (addr.ss_family != AF_UNSPEC)
		return 1;
	return -1;
}

/*
 * Sub-process performing lookup
 */
void
do_query(int fd, int family, const char *name)
{
	struct sockaddr_storage ss;
	struct addrinfo hints, *res;
	int	r, c;

	memset(&hints, 0, sizeof(hints));
	hints.ai_socktype = SOCK_DGRAM;
#ifdef notyet
	/* XXX - specify AI_whatever for automatic v4-v6 mapping */
#else
	hints.ai_family = (family == AF_INET)? family : AF_UNSPEC;
#endif

	r = getaddrinfo(name, "syslog", &hints, &res);
	if (r != 0) {
		dprintf("failed to resolve \"%s\": %s\n",
				name, gai_strerror(r));
		exit(1);
	}

	memset(&ss, 0, sizeof(ss));

	/* Do v4mapped v6 address (supported by recent glibcs
	 * only) */
	if (family == AF_INET6 && res->ai_family == AF_INET) {
		struct sockaddr_in6 *six = (struct sockaddr_in6 *) &ss;
		struct sockaddr_in  *sin = (struct sockaddr_in *) res->ai_addr;

		six->sin6_family = AF_INET6;
		six->sin6_port = sin->sin_port;
		six->sin6_addr.s6_addr16[5] = 0xffff;
		six->sin6_addr.s6_addr32[3] = sin->sin_addr.s_addr;
	} else if (res->ai_addrlen > sizeof(ss)) {
		dprintf("failed to resolve \"%s\": address too large\n",
				name);
		exit(1);
	} else {
		memcpy(&ss, res->ai_addr, res->ai_addrlen);
	}

	c = 100;
	do {
		if ((r = write(fd, &ss, sizeof(ss))) < 0) {
			if ((errno != EAGAIN) && (errno != EINTR)) {
				dprintf("failed to write: %s\n", strerror(errno));
				exit(1);
			}
			usleep(50000);
		}
	} while ((r < 0) && (c-- > 0));

	if (r >= 0)
		exit(0);
	dprintf("failed to write: %s\n", strerror(errno));
	exit(1);
}
